import numpy as np
from lmfit import Parameters, minimize


def _heartrate_model_predict(model_params, power):
    power = power + power.cumsum() * model_params["hr_drift"]
    hr_lin = power * model_params["dhr"] + model_params["hr_rest"]
    hr_ss = np.minimum(model_params["hr_max"], hr_lin)

    dhr = np.insert(np.diff(power), 0, 0)
    tau = np.where((dhr <= 0), model_params["tau_fall"], model_params["tau_rise"])

    predicted_hr = []
    hr_previous = model_params["hr_rest"]
    for h, t in zip(hr_ss, tau):
        hr_previous = hr_previous + (h - hr_previous) / t
        predicted_hr.append(hr_previous)

    return predicted_hr


def _heartrate_model_residuals(model_params, power, heartrate):
    model = _heartrate_model_predict(model_params, power)
    return heartrate - model


def heartrate_model(heartrate, power, **kwargs):
    """
    Source:
    De Smet et al., Heart rate modelling as a potential physical finess assessment for runners and cyclists.
    http://ceur-ws.org/Vol-1842/paper_13.pdf
    """
    # Initial model parameters
    model_params = Parameters()
    model_params.add_many(
        ("hr_rest", kwargs.get("hr_rest", 75)),
        ("hr_max", kwargs.get("hr_max", 200)),
        ("dhr", kwargs.get("dhr", 0.30)),
        ("tau_rise", kwargs.get("tau_rise", 24)),
        ("tau_fall", kwargs.get("tau_fall", 30)),
        ("hr_drift", kwargs.get("hr_drift", 3 * 10 ** -5)),
    )

    model = minimize(
        fcn=_heartrate_model_residuals,
        params=model_params,
        method="nelder",  # Nelder-Mead
        args=(power, heartrate),
    )

    predictions = _heartrate_model_predict(model.params, power)

    return model, predictions



import math

import numpy as np
import pandas as pd


def tau_w_prime_balance(power, cp, untill=None):
    if untill is None:
        untill = len(power)

    avg_power_below_cp = power[:untill][power[:untill] < cp].mean()
    if math.isnan(avg_power_below_cp):
        avg_power_below_cp = 0
    delta_cp = cp - avg_power_below_cp

    return 546 * math.e ** (-0.01 * delta_cp) + 316


def get_tau_method(power, cp, tau_dynamic, tau_value):
    if tau_dynamic:
        tau_dynamic = [tau_w_prime_balance(power, cp, i) for i in range(len(power))]
        tau = lambda t: tau_dynamic[t]

    elif tau_value is None:
        static_tau = tau_w_prime_balance(power, cp)
        tau = lambda t: static_tau

    else:
        tau = lambda t: tau_value

    return tau


def w_prime_balance_waterworth(
    power, cp, w_prime, tau_dynamic=False, tau_value=None, *args, **kwargs
):
    """
    Optimisation of Skiba's algorithm by Dave Waterworth.
    Source:
    http://markliversedge.blogspot.nl/2014/10/wbal-optimisation-by-mathematician.html
    Source:
    Skiba, Philip Friere, et al. "Modeling the expenditure and reconstitution of work capacity above critical power." Medicine and science in sports and exercise 44.8 (2012): 1526-1532.
    """
    sampling_rate = 1
    running_sum = 0
    w_prime_balance = []
    tau = get_tau_method(power, cp, tau_dynamic, tau_value)

    for t, p in enumerate(power):
        power_above_cp = p - cp
        w_prime_expended = max(0, power_above_cp) * sampling_rate
        running_sum = running_sum + w_prime_expended * (
            math.e ** (t * sampling_rate / tau(t))
        )

        w_prime_balance.append(
            w_prime - running_sum * math.e ** (-t * sampling_rate / tau(t))
        )

    return pd.Series(w_prime_balance)


def w_prime_balance_skiba(
    power, cp, w_prime, tau_dynamic=False, tau_value=None, *args, **kwargs
):
    """
    Source:
    Skiba, Philip Friere, et al. "Modeling the expenditure and reconstitution of work capacity above critical power." Medicine and science in sports and exercise 44.8 (2012): 1526-1532.
    """
    w_prime_balance = []
    tau = get_tau_method(power, cp, tau_dynamic, tau_value)

    for t in range(len(power)):
        w_prime_exp_sum = 0

        for u, p in enumerate(power[: t + 1]):
            w_prime_exp = max(0, p - cp)
            w_prime_exp_sum += w_prime_exp * np.power(np.e, (u - t) / tau(t))

        w_prime_balance.append(w_prime - w_prime_exp_sum)

    return pd.Series(w_prime_balance)


def w_prime_balance_froncioni_skiba_clarke(power, cp, w_prime):
    """
    Source:
    Skiba, P. F., Fulford, J., Clarke, D. C., Vanhatalo, A., & Jones, A. M. (2015). Intramuscular determinants of the ability to recover work capacity above critical power. European journal of applied physiology, 115(4), 703-713.
    """
    last = w_prime
    w_prime_balance = []

    for p in power:
        if p < cp:
            new = last + (cp - p) * (w_prime - last) / w_prime
        else:
            new = last + (cp - p)

        w_prime_balance.append(new)
        last = new

    return pd.Series(w_prime_balance)


def w_prime_balance(power, cp, w_prime, algorithm="waterworth", *args, **kwargs):
    if algorithm == "waterworth":
        method = w_prime_balance_waterworth
    elif algorithm == "skiba":
        method = w_prime_balance_skiba
    elif algorithm == "froncioni-skiba-clarke":
        method = w_prime_balance_froncioni_skiba_clarke

    return method(power, cp, w_prime, *args, **kwargs)



from enum import Enum


class DataTypeEnum(str, Enum):
    latitude = "latitude"
    longitude = "longitude"
    speed = "speed"
    power = "power"
    left_right_balance = "left-right balance"
    elevation = "elevation"
    cadence = "cadence"
    heartrate = "heartrate"
    temperature = "temperature"
    distance = "distance"



from datetime import timedelta

import pandas as pd
from stravalib.client import Client

from .utils import resample_data


STREAM_TYPES = [
    "time",
    "latlng",
    "distance",
    "altitude",
    "velocity_smooth",
    "heartrate",
    "cadence",
    "watts",
    "temp",
    "moving",
    "grade_smooth",
]

COLUMN_TRANSLATIONS = {
    "altitude": "elevation",
    "velocity_smooth": "speed",
    "watts": "power",
    "temp": "temperature",
    "grade_smooth": "grade",
}


def read_strava(
    activity_id: int,
    access_token: str,
    refresh_token: str = None,
    client_id: int = None,
    client_secret: str = None,
    resample: bool = False,
    interpolate: bool = False,
) -> pd.DataFrame:
    """This method lets you retrieve activity data from Strava.
    Columns names are translated to sweat terminology (e.g. "heart_rate" > "heartrate").
    Two API calls are made to the Strava API: 1 to retrieve activity metadata, 1 to retrieve the raw data ("streams").

    Args:
        activity_id: The id of the activity
        access_token: The Strava access token
        refresh_token: The Strava refresh token. Optional.
        client_id: The Strava client id. Optional. Used for token refresh.
        client_secret: The Strava client secret. Optional. Used for token refresh.
        resample: whether or not the data frame needs to be resampled to 1Hz
        interpolate: whether or not missing data in the data frame needs to be interpolated

    Returns:
        A pandas data frame with all the data.
    """
    client = Client()
    client.access_token = access_token
    client.refresh_token = refresh_token

    activity = client.get_activity(activity_id)
    start_datetime = activity.start_date_local

    streams = client.get_activity_streams(
        activity_id=activity_id,
        types=STREAM_TYPES,
        series_type="time",
    )

    raw_data = dict()
    for key, value in streams.items():
        if key == "latlng":
            latitude, longitude = list(zip(*value.data))
            raw_data["latitude"] = latitude
            raw_data["longitude"] = longitude
        else:
            try:
                key = COLUMN_TRANSLATIONS[key]
            except KeyError:
                pass

            raw_data[key] = value.data

    data = pd.DataFrame(raw_data)

    def time_to_datetime(time):
        return start_datetime + timedelta(seconds=time)

    data["datetime"] = data["time"].apply(time_to_datetime)

    data = data.drop(["time"], axis="columns")

    data = data.set_index("datetime")

    data = resample_data(data, resample, interpolate)

    return data



import xml.etree.ElementTree as ET
from pathlib import Path

import pandas as pd

from .utils import remove_duplicate_indices, resample_data


NAMESPACES = {
    "default": "http://www.topografix.com/GPX/1/1",
    "gpxtpx": "http://www.garmin.com/xmlschemas/TrackPointExtension/v1",
    "gpxx": "http://www.garmin.com/xmlschemas/GpxExtensions/v3",
    "xsi": "http://www.w3.org/2001/XMLSchema-instance",
}


def xml_find_value_or_none(element, match, namespaces=None):
    e = element.find(match, namespaces=namespaces)
    if e is None:
        return e
    else:
        return e.text


def read_gpx(fpath, resample: bool = False, interpolate: bool = False) -> pd.DataFrame:
    """This method loads a GPX file into a Pandas DataFrame.
    Columns names are translated to sweat terminology (e.g. "heart_rate" > "heartrate").

    Args:
        fpath: str, file-like or Path object
        resample: whether or not the data frame needs to be resampled to 1Hz
        interpolate: whether or not missing data in the data frame needs to be interpolated

    Returns:
        A pandas data frame with all the data.
    """
    tree = ET.parse(Path(fpath))
    root = tree.getroot()
    trk = root.find("default:trk", NAMESPACES)
    trkseg = trk.find("default:trkseg", NAMESPACES)

    records = []
    for trackpoint in trkseg.findall("default:trkpt", NAMESPACES):
        latitude = trackpoint.attrib.get("lat", None)
        longitude = trackpoint.attrib.get("lon", None)

        elevation = xml_find_value_or_none(trackpoint, "default:ele", NAMESPACES)

        datetime = xml_find_value_or_none(trackpoint, "default:time", NAMESPACES)

        extensions = trackpoint.find("default:extensions", NAMESPACES)

        power = xml_find_value_or_none(extensions, "default:power", NAMESPACES)

        trackpoint_extension = extensions.find("gpxtpx:TrackPointExtension", NAMESPACES)

        temperature = xml_find_value_or_none(
            trackpoint_extension, "gpxtpx:atemp", NAMESPACES
        )
        heartrate = xml_find_value_or_none(
            trackpoint_extension, "gpxtpx:hr", NAMESPACES
        )
        cadence = xml_find_value_or_none(trackpoint_extension, "gpxtpx:cad", NAMESPACES)

        records.append(
            dict(
                latitude=pd.to_numeric(latitude),
                longitude=pd.to_numeric(longitude),
                elevation=pd.to_numeric(elevation),
                datetime=datetime,
                power=pd.to_numeric(power),
                temperature=pd.to_numeric(temperature),
                heartrate=pd.to_numeric(heartrate),
                cadence=pd.to_numeric(cadence),
            )
        )

    gpx_df = pd.DataFrame(records)
    gpx_df = gpx_df.dropna("columns", "all")
    gpx_df["datetime"] = pd.to_datetime(gpx_df["datetime"], utc=True)
    gpx_df = gpx_df.set_index("datetime")

    gpx_df = remove_duplicate_indices(gpx_df)

    gpx_df = resample_data(gpx_df, resample, interpolate)

    return gpx_df



from enum import Enum
from pathlib import Path
from typing import List

from yaml import load, Loader
from pydantic import BaseModel, parse_obj_as, validator

from sweat.constants import DataTypeEnum


class SportEnum(str, Enum):
    cycling = "cycling"
    running = "running"
    swimming = "swimming"


class FileTypeEnum(str, Enum):
    fit = "FIT"
    tcx = "TCX"
    gpx = "GPX"


class Sensor(BaseModel):
    name: str
    data_types: List[DataTypeEnum]


class ExampleData(BaseModel):
    description: str = None
    path: Path
    sport: SportEnum
    file_type: FileTypeEnum
    recording_device: str
    recording_device: str
    sensors: List[Sensor] = None
    included_data: List[DataTypeEnum]
    laps: int = None
    sessions: int = None

    @validator("path")
    def make_path_absolute(cls, v):
        return Path(examples_data_dir(), v)


def examples_dir():
    return Path(".", __file__).parent


def examples_data_dir():
    return Path(examples_dir(), "data")


def examples(*, path=None, file_type=None, sport=None):
    if path is not None and (file_type is not None or sport is not None):
        raise ValueError(
            "Providing both the path argument and one or more of file_type and sport is not supported."
        )

    with Path(examples_dir(), "index.yml").open("r") as f:
        index = load(f, Loader=Loader)
    loaded_data = parse_obj_as(List[ExampleData], index)

    if path is not None:
        example_data_dir = examples_data_dir()
        path = Path(example_data_dir, path)
        for example in examples():
            if example.path == path:
                return example
        raise ValueError(f'Example data "{path}" not found.')

    if file_type is not None:
        loaded_data = filter(lambda d: d.file_type == file_type, examples())

    if sport is not None:
        loaded_data = filter(lambda d: d.sport == sport, examples())

    return loaded_data



from . import pandas
from .examples.utils import examples, FileTypeEnum, SportEnum
from .io.fit import read_fit
from .io.gpx import read_gpx
from .io.generic import read_dir, read_file
from .io.strava import read_strava
from .io.tcx import read_tcx
from .pdm.regressors import PowerDurationRegressor
from .pdm.w_prime_balance import w_prime_balance
from .utils import array_1d_to_2d



import xml.etree.ElementTree as ET
from pathlib import Path

import pandas as pd

from .utils import remove_duplicate_indices, resample_data


NAMESPACES = {
    "default": "http://www.garmin.com/xmlschemas/TrainingCenterDatabase/v2",
    "ns5": "http://www.garmin.com/xmlschemas/ActivityGoals/v1",
    "ns3": "http://www.garmin.com/xmlschemas/ActivityExtension/v2",
    "ns2": "http://www.garmin.com/xmlschemas/UserProfile/v2",
}


def xml_find_value_or_none(element, match, namespaces=None):
    e = element.find(match, namespaces=namespaces)
    if e is None:
        return e
    else:
        return e.text


def read_tcx(fpath, resample: bool = False, interpolate: bool = False) -> pd.DataFrame:
    """This method loads a TCX file into a Pandas DataFrame.
    Columns names are translated to sweat terminology (e.g. "heart_rate" > "heartrate").

    Args:
        fpath: str, file-like or Path object
        resample: whether or not the data frame needs to be resampled to 1Hz
        interpolate: whether or not missing data in the data frame needs to be interpolated

    Returns:
        A pandas data frame with all the data.
    """
    tree = ET.parse(Path(fpath))
    root = tree.getroot()
    activities = root.find("default:Activities", NAMESPACES)

    records = []
    lap_no = -1
    session = 0
    for activity in activities.findall("default:Activity", NAMESPACES):
        for lap in activity.findall("default:Lap", NAMESPACES):
            lap_no += 1
            track = lap.find("default:Track", NAMESPACES)
            for trackpoint in track.findall("default:Trackpoint", NAMESPACES):
                datetime = xml_find_value_or_none(
                    trackpoint, "default:Time", NAMESPACES
                )
                elevation = xml_find_value_or_none(
                    trackpoint, "default:AltitudeMeters", NAMESPACES
                )
                distance = xml_find_value_or_none(
                    trackpoint, "default:DistanceMeters", NAMESPACES
                )
                cadence = xml_find_value_or_none(
                    trackpoint, "default:Cadence", NAMESPACES
                )

                position = trackpoint.find("default:Position", NAMESPACES)
                latitude = xml_find_value_or_none(
                    position, "default:LatitudeDegrees", NAMESPACES
                )
                longitude = xml_find_value_or_none(
                    position, "default:LongitudeDegrees", NAMESPACES
                )

                hr = trackpoint.find("default:HeartRateBpm", NAMESPACES)
                heartrate = xml_find_value_or_none(hr, "default:Value", NAMESPACES)

                extensions = trackpoint.find("default:Extensions", NAMESPACES)
                if extensions:
                    tpx = extensions.find("ns3:TPX", NAMESPACES)
                    speed = xml_find_value_or_none(tpx, "ns3:Speed", NAMESPACES)
                    power = xml_find_value_or_none(tpx, "ns3:Watts", NAMESPACES)
                else:
                    speed = None
                    power = None

                records.append(
                    dict(
                        datetime=datetime,
                        latitude=latitude,
                        longitude=longitude,
                        elevation=elevation,
                        heartrate=heartrate,
                        cadence=cadence,
                        distance=distance,
                        speed=speed,
                        power=power,
                        lap=lap_no,
                        session=session,
                    )
                )

    tcx_df = pd.DataFrame(records)
    tcx_df = tcx_df.dropna("columns", "all")
    tcx_df["datetime"] = pd.to_datetime(tcx_df["datetime"], utc=True)
    tcx_df = tcx_df.set_index("datetime")

    # Convert columns to numeric if possible
    tcx_df = tcx_df.apply(pd.to_numeric, errors="ignore")

    tcx_df = remove_duplicate_indices(tcx_df)

    tcx_df = resample_data(tcx_df, resample, interpolate)

    return tcx_df



import numpy as np
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from scipy.optimize import curve_fit


class PowerDurationRegressor(BaseEstimator, RegressorMixin):
    """Based on: https://scikit-learn.org/stable/developers/develop.html"""

    def __init__(
        self,
        model="2 param",
        cp=300,
        w_prime=20000,
        p_max=1000,
        tau=300,
        tcp_max=1800,
        a=50,
    ):
        self.model = model
        self.cp = cp
        self.w_prime = w_prime
        self.p_max = p_max
        self.tau = tau
        self.tcp_max = tcp_max
        self.a = a

    def _2_param_model(self, X, cp, w_prime, *args):
        t = X.T[0]  # X should be a (1, n) array
        return cp + w_prime / t

    def _3_param_model(self, X, cp, w_prime, p_max, *args):
        t = X.T[0]  # X should be a (1, n) array
        return w_prime / (t + (w_prime / (p_max - cp))) + cp

    def _exp_model(self, X, cp, p_max, tau, *args):
        """
        Source: Hopkins, W. G., Edmond, I. M., Hamilton, B. H., Macfarlane, D. J., & Ross, B. H. (1989). Relation between power and endurance for treadmill running of short duration. Ergonomics, 32(12), 1565-1571.
        """
        t = X.T[0]  # X should be a (1, n) array
        return (p_max - cp) * np.exp(-t / tau) + cp

    def _omni_model(self, X, cp, p_max, w_prime, a, *args):
        """
        Source: Puchowicz, M. J., Baker, J., & Clarke, D. C. (2020). Development and field validation of an omni-domain power-duration model. Journal of Sports Sciences, 38(7), 801-813.
        """
        t = X.T[0]  # X should be a (1, n) array

        result = w_prime / t * (1 - np.exp(-t * (p_max - cp) / w_prime)) + cp

        return np.where(
            t <= self.tcp_max, result, result - a * np.log(t / self.tcp_max)
        )

    def _model_selection(self):
        if self.model == "2 param":
            func = self._2_param_model
            params = ["cp", "w_prime"]
            return func, params
        elif self.model == "3 param":
            func = self._3_param_model
            params = ["cp", "w_prime", "p_max"]
            return func, params
        elif self.model == "exponential":
            func = self._exp_model
            params = ["cp", "p_max", "tau"]
            return func, params
        elif self.model == "omni":
            func = self._omni_model
            params = ["cp", "p_max", "w_prime", "a"]
            return func, params
        else:
            raise ValueError(f"self.model has an invalid value: {self.model}")

    def fit(self, X, y):
        self.n_features_in_ = 1
        X, y = check_X_y(X, y, ensure_min_samples=2)

        func, params = self._model_selection()

        initial_params = [getattr(self, param_name) for param_name in params]
        fitted_params, _ = curve_fit(f=func, xdata=X, ydata=y, p0=initial_params)

        for name, value in zip(params, fitted_params):
            setattr(self, f"{name}_", value)

        self.is_fitted_ = True

        return self

    def predict(self, X):
        check_is_fitted(self)
        X = check_array(X)

        func, params = self._model_selection()

        args = []
        for param_name in params:
            args.append(getattr(self, f"{param_name}_"))

        return func(X, *args)

    def _more_tags(self):
        return {"poor_score": True}



import numpy as np
import pandas as pd
from collections import namedtuple


def mask_fill(y, mask=None, value=0.0):
    """Replace masked values

    Parameters
    ----------
    y : ndarray
    mask : array-like of bools, optional
        Default value is None, which means no masking will be applied
    value : number, optional
        Value to use for replacement, default=0.0

    Returns
    -------
    y: ndarray

    All operations will be performed on the original array.
    To preserve original array pass a copy to the function
    """

    if mask is None:
        return y

    mask = np.array(mask, dtype=bool)
    y[~mask] = value

    return y


def rolling_mean(y, window=10, mask=None, algorithm="uniform", value=0.0):
    """Compute rolling mean

    Compute *uniform* or *ewma* rolling mean of the stream. In-process masking with replacement is
    controlled by optional keyword parameters

    Parameters
    ----------
    y : ndarray
    window : int
        Size of the moving window in sec, default=10
    mask : array-like of boolean, optional
        Default value is None, which means no masking will be applied
    algorithm : {"uniform", "emwa"}, optional
        Type of averaging, default="uniform"
    value : number, optional
        Value to use for replacement, default=0.0

    Returns
    -------
    y: ndarray

    The moving array will indicate which samples to set to zero before
    applying rolling mean.
    """
    if mask is not None:
        y = mask_fill(y, mask, value)

    y = pd.Series(y)

    if algorithm == "ewma":
        y = y.ewm(span=window, min_periods=1).mean().values
    elif algorithm == "uniform":
        y = y.rolling(window, min_periods=1).mean().values

    return y


def median_filter(y, window=31, threshold=1, value=None):
    """Outlier replacement using median filter

    Detect outliers using median filter and replace with rolling median or specified value

    Parameters
    ----------
    y : ndarray
    window : int, optional
        Size of window (including the sample; default=31 is equal to 15 on either side of value)
    threshold : number, optional
        default=3 and corresponds to 2xSigma
    value : float, optional
        Value to be used for replacement, default=None, which means replacement by rolling median value

    Returns
    -------
    y: ndarray

    All operations will be performed on the original array.
    To preserve original array pass a copy to the function
    """
    y = pd.Series(y)

    rolling_median = y.rolling(window, min_periods=1).median()

    difference = np.abs(y - rolling_median)

    median_abs_deviation = difference.rolling(window, min_periods=1).median()

    outlier_idx = difference > 1.4826 * threshold * median_abs_deviation
    """ The factor 1.4826 makes the MAD scale estimate
        an unbiased estimate of the standard deviation for Gaussian data.
    """

    if value:
        y[outlier_idx] = value
    else:
        y[outlier_idx] = rolling_median[outlier_idx]

    return y.values


# FTP based 7-zones with left bind edge set to -0.001
POWER_ZONES_THRESHOLD = [-0.001, 0.55, 0.75, 0.9, 1.05, 1.2, 1.5, 10.0]
POWER_ZONES_THRESHOLD_DESC = [
    "Active Recovery",
    "Endurance",
    "Tempo",
    "Threshold",
    "VO2Max",
    "Anaerobic",
    "Neuromuscular",
]
POWER_ZONES_THRESHOLD_ZNAME = ["Z1", "Z2", "Z3", "Z4", "Z5", "Z6", "Z7"]

# LTHR based 5-zones with left bind edge set to -0.001
HEART_RATE_ZONES = [-0.001, 0.68, 0.83, 0.94, 1.05, 10.0]
HEART_RATE_ZONES_DESC = [
    "Active recovery",
    "Endurance",
    "Tempo",
    "Threshold",
    "VO2Max",
]
HEART_RATE_ZONES_ZNAME = ["Z1", "Z2", "Z3", "Z4", "Z5"]


def compute_zones(y, zones=None, ftp=None, lthr=None, labels=None):
    """Convert stream into respective zones stream

    Watts streams can be converted either into ftp based 7-zones or into custom zones
    HR streams can be converted either in lthr based 5-zones or into custom zones
    One of three *ftp*, *lthr* or *zone* keyword parameters must be provided

    Parameters
    ----------
    y : ndarray
    ftp : number, optional
        Value for FTP, will be used for 7-zones calculation
    lthr: number, optional
        Value for LTHR, will be used for 5-zones calculation
    zones: list, optional
        List of custom defined zones with left edge set to -1 and right edge to 10000

    Returns
    -------
    array-like of int, the same type as arg
    """

    if zones is not None:
        abs_zones = zones
    elif ftp is not None:
        abs_zones = np.asarray(POWER_ZONES_THRESHOLD) * ftp
    elif lthr is not None:
        abs_zones = np.asarray(HEART_RATE_ZONES) * lthr
    else:
        raise ValueError(
            "One of the keyword arguments [zones, ftp, lthr] must be provided"
        )

    if labels is None:
        labels = list(range(1, len(abs_zones)))
    assert len(abs_zones) == (len(labels) + 1)

    y = pd.cut(y, bins=abs_zones, labels=labels)

    return y


def best_interval(y, window, mask=None, value=0.0):
    """Compute best interval of the stream

    Masking with replacement is controlled by keyword arguments

    Parameters
    ----------
    y: ndarray
    window : int
        Duration of the interval in seconds
    mask : array-like of bool, optional
        default=None, which means no masking
    value : number, optional
        Value to use for replacement, default=0.0

    Returns
    -------
    float
    """

    y = rolling_mean(y, window=window, mask=mask, value=value)

    rv = np.max(y)

    return rv


def time_in_zones(y, **kwargs):
    """Time in zones

    Calculate time [sec] spent in each zone

    Parameters
    ----------
    y : ndarray, power or heartrate
    kwargs : see zones

    Returns
    -------
    ndarray
    """
    z = pd.Series(compute_zones(y, **kwargs))
    tiz = z.groupby(z).count()

    return tiz.values


def weighted_average_power(y, mask=None, algorithm="WAP", value=0.0):
    """Weighted average power

    Parameters
    ----------
    y : ndarray
        Power stream
    mask: array-like of bool, optional
        default=None, which means no masking
    algorithm : {"WAP", "xPower"}
        Determines calculation method to use, default='WAP'
    value : number, optional
        Value to use for replacement, default=0.0

    Returns
    -------
    number
    """

    if algorithm == "WAP":
        _rolling_mean = rolling_mean(y, window=30, mask=mask, value=value)
    elif algorithm == "xPower":
        _rolling_mean = rolling_mean(
            y, window=25, mask=mask, value=value, algorithm="emwa"
        )

    rv = np.mean(np.power(_rolling_mean, 4)) ** (1 / 4)

    return rv


def mean_max(y, mask=None, value=0.0, monotonic=False):
    """Mean-max curve

    Compute mean-max (power duration curve) from the stream. Mask-filter options can be
    added using the keyword arguments.

    Parameters
    ----------
    arg : array-like
        Power stream
    mask: array-like, optional
        Replacement mask (the default is None, which implies no masking)
    value: number, optional
        Value to use as a replacement (the default is 0.0)

    Returns
    -------
    ndarray
    """

    y = mask_fill(y, mask=mask, value=value)
    y = pd.Series(y)

    # Compute the accumulated energy from the power data
    energy = y.cumsum()

    # Compute the maximum sustainable power using the difference in energy
    # This method is x4 faster than using rolling mean
    y = np.array([])
    for t in np.arange(1, len(energy)):
        y = np.append(y, energy.diff(t).max() / (t))

    if monotonic:
        monotonic_y = []
        previous = y[-1]
        for p in np.flip(y):
            if p <= previous:
                monotonic_y.append(previous)
            else:
                monotonic_y.append(p)
                previous = p

        y = np.flip(monotonic_y)

    return y


DataPoint = namedtuple("DataPoint", ["index", "value"])


def multiple_best_intervals(arg, duration, number):
    """Compute multiple best intervals

    TODO: This function should return a list of {'start_index': v, 'stop_index': v, 'index': v, 'value': v}
    Parameters
    ----------
    arg : pd.Stream
    duration : number
    number : int

    Returns
    -------
    ndarray
    """
    moving_average = arg.rolling(duration).mean()
    length = len(moving_average)
    mean_max_bests = []

    for i in range(number):
        if moving_average.isnull().all():
            mean_max_bests.append(DataPoint(np.nan, np.nan))
            continue

        max_value = moving_average.max()
        max_index = moving_average.idxmax()
        mean_max_bests.append(DataPoint(max_index, max_value))

        # Set moving averages that overlap with last found max to np.nan
        overlap_min_index = max(0, max_index - duration)
        overlap_max_index = min(length, max_index + duration)
        moving_average.loc[overlap_min_index:overlap_max_index] = np.nan

    return mean_max_bests



import functools
import inspect
import sys
import types

import numpy as np
import pandas as pd


CAST_TYPES = [list, pd.Series]


def type_casting(func):
    """Type casting
    This decorator casts input arguments of types [list, pandas.Series] to numpy.ndarray
    so the algorithms accept these input arguments.
    As a bonus, the decorator casts the return value of the algorithm to the type of the first
    array-like input argument.

    Parameters
    ----------
    module_or_func : [types.ModuleType, types.FunctionType]
        Module or function that has to be type casted. Can be None.

    Returns
    -------
    function
        Decorated function.
    """

    @functools.wraps(func)
    def func_wrapper(*args, **kwargs):
        output_type = None
        new_args = []

        for arg in args:
            input_type = type(arg)

            if input_type in CAST_TYPES:
                new_args.append(np.asarray(arg))

                if output_type is None:
                    # Type of first array-like argument is used for output casting
                    output_type = input_type
            else:
                new_args.append(arg)

        # There is no use-case for type casting kwargs now.
        # If there is, this can be uncommented and tests can be added.
        # new_kwargs = dict()
        # for key, value in kwargs.items():
        #     input_type = type(value)

        #     if input_type in CAST_TYPES:
        #         new_kwargs[key] = np.asarray(value)

        #         if output_type is None:
        #             # Type of first array-like argument is used for output casting
        #             output_type = input_type
        #     else:
        #         new_kwargs[key] = value
        #
        # output = func(*new_args, **new_kwargs)

        output = func(*new_args)
        if output_type is not None and isinstance(output, np.ndarray):
            return output_type(output)
        else:
            return output

    return func_wrapper


def enable_type_casting(module_or_func=None):
    """Enable type casting
    This method enables casting of input arguments to numpy.ndarray so the algorithms accept
    array-like input arguments of types list and pandas.Series.
    As a bonus, the return value of the algorithm is casted to the type of the first array-like input argument.

    Parameters
    ----------
    module_or_func : [types.ModuleType, types.FunctionType]
        Module or function that has to be type casted. Can be None.

    Returns
    -------
    function
        Decorated function.
    """
    if module_or_func is None:
        # Because sys.modules changes during this operation we cannot loop over sys.modules directly
        key_values = [(key, value) for key, value in sys.modules.items()]
        for key, value in key_values:
            # @TODO this if statement might not cover all cases (or too much cases)
            if (
                key.startswith("sweat.hrm")
                or key.startswith("sweat.pdm")
                or key.startswith("sweat.metrics")
            ):
                enable_type_casting(module_or_func=value)

    elif isinstance(module_or_func, types.ModuleType):
        for name, obj in [
            (name, obj) for name, obj in inspect.getmembers(module_or_func)
        ]:
            if (
                inspect.isfunction(obj)
                and inspect.getmodule(obj).__package__ == module_or_func.__package__
            ):
                func = getattr(module_or_func, name)
                setattr(module_or_func, name, type_casting(func))

    elif isinstance(module_or_func, types.FunctionType):
        return type_casting(module_or_func)

    else:
        raise ValueError(
            "enable_type_casting takes arguments of types [ModuleType, FunctionType]"
        )


def array_1d_to_2d(l):
    return np.asarray(l).reshape(-1, 1)



import pandas as pd


def wpk(power, weight):
    """Watts per kilo

    Parameters
    ----------
    power : ndarray
    weight : number

    Returns
    -------
    array-like
    """

    return power / weight


def relative_intensity(wap, threshold_power):
    """Relative intensity

    Parameters
    ----------
    wap : number
        WAP or xPower
    threshold_power : number
        FTP or CP

    Returns
    -------
    float
        IF or RI
    """

    rv = wap / threshold_power

    return rv


def stress_score(wap, threshold_power, duration):
    """Stress Score

    Parameters
    ----------
    wap : number
        WAP or xPower
    threshold_power : number
        FTP or CP
    duration : int
        Duration in seconds

    Returns
    -------
    ss
    """

    ss = (duration / 3600) * (wap / threshold_power) ** 2 * 100

    return ss



from typing import List, Union

import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype

from .constants import DataTypeEnum
from .metrics import core


@pd.api.extensions.register_dataframe_accessor("sweat")
class SweatAccessor:
    def __init__(self, pandas_obj):
        self._validate(pandas_obj)
        self._obj = pandas_obj

    @staticmethod
    def _validate(obj):
        if not isinstance(obj.index, (pd.DatetimeIndex, pd.TimedeltaIndex)):
            raise AttributeError("DataFrame Index should be a DateTimeIndex.")

        if not all(obj.index.to_series().diff()[1:] == np.timedelta64(1, "s")):
            raise AttributeError(
                "Data is not sampled at a regular 1s interval. Consider resampling first."
            )

        for data_type in DataTypeEnum:
            # @TODO figure out which columns need to be numeric continue
            continue
            try:
                if not is_numeric_dtype(obj[data_type.value]):
                    raise AttributeError(f"Column {data_type.value} is not numeric")
            except KeyError:
                continue

    def mean_max(
        self, columns: Union[List, str], monotonic: bool = False
    ) -> pd.DataFrame:
        if isinstance(columns, str):
            columns = [columns]

        data = None
        for column in columns:
            result = core.mean_max(self._obj[column], monotonic=monotonic)

            if data is None:
                index = pd.to_timedelta(range(len(result)), unit="s")
                data = pd.DataFrame(index=index)

            data["mean_max_" + column] = result

        return data

    def to_timedelta_index(self):
        """This method converts the index to a relative TimedeltaIndex, returning a copy of the data frame with the new index.

        Returns:
            A pandas data frame with a TimedeltaIndex.
        """
        return self._obj.set_index(self._obj.index - self._obj.index[0])


@pd.api.extensions.register_series_accessor("sweat")
class SweatSeriesAccessor:
    def __init__(self, pandas_obj):
        self._validate(pandas_obj)
        self._obj = pandas_obj

    @staticmethod
    def _validate(obj):
        if not isinstance(obj.index, pd.DatetimeIndex):
            raise AttributeError("DataFrame Index should be a DateTimeIndex.")

        if not all(obj.index.to_series().diff()[1:] == np.timedelta64(1, "s")):
            raise AttributeError(
                "Data is not sampled at a regular 1s interval. Consider resampling first."
            )

        if not is_numeric_dtype(obj):
            raise AttributeError(f"Series dtype should be numeric")

    def mean_max(self, monotonic: bool = False) -> pd.Series:
        """This method calculates the mean max values of the series.

        Returns:
            A pandas series with a TimedeltaIndex.
        """
        result = core.mean_max(self._obj, monotonic=monotonic)
        index = pd.to_timedelta(range(len(result)), unit="s")
        return pd.Series(result, index=index, name="mean_max_" + self._obj.name)

    def to_timedelta_index(self):
        """This method converts the index to a relative TimedeltaIndex, returning a copy of the series with the new index.

        Returns:
            A pandas series with a TimedeltaIndex.
        """
        index = self._obj.index - self._obj.index[0]
        return pd.Series(self._obj, index=index)

    def calculate_zones(
        self, bins: List[int], labels: List[Union[int, str]]
    ) -> pd.Series:
        """Returns a pandas.Series with the zone label for each value.
        This method uses the pandas.cut() method under the hood.
        Nan will be returned for values that are not in any of the bins.

        Args:
            bins: Left and right bounds for each zone.
            labels: Specifies the labels for the zones. Must be the same length as the resulting zones.

        Returns:
            A pandas series with the zone label for each value.
        """
        return pd.cut(self._obj, bins=bins, labels=labels)

    def time_in_zone(self, bins: List[int], labels: List[Union[int, str]]) -> pd.Series:
        """Returns a pandas.Series with the value counts for each zone.
        This method uses the pandas.Series.value_counts() method under the hood.

        Args:
            bins: Left and right bounds for each zone.
            labels: Specifies the labels for the zones. Must be the same length as the resulting zones.

        Returns:
            A pandas series with the value counts for each zones.
        """
        zones = self.calculate_zones(bins, labels)
        zones_value_counts = zones.value_counts()
        return pd.to_timedelta(zones_value_counts, unit="s")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sweat',
 'sweat.examples',
 'sweat.hrm',
 'sweat.io',
 'sweat.metrics',
 'sweat.pdm']

package_data = \
{'': ['*'], 'sweat.examples': ['data/*']}

install_requires = \
['fitparse>=1.1.0,<2.0.0',
 'lmfit>=1.0.0,<2.0.0',
 'pandas>=1.0.3,<2.0.0',
 'pydantic>=1.4,<2.0',
 'pyyaml>=5.3.1,<6.0.0',
 'requests>=2.23.0,<3.0.0',
 'scikit-learn>=0.23.1,<0.24.0',
 'scipy>=1.4.1,<2.0.0',
 'stravalib>=0.10.2,<0.11.0']

setup_kwargs = {
    'name': 'sweat',
    'version': '0.18.0',
    'description': 'Workout analysis',
    'long_description': '# Sweatpy\n\n[![Downloads](https://pepy.tech/badge/sweat)](https://pepy.tech/project/sweat)\n\n> :warning: **Sweatpy is currently undergoing major revisions which will result in deprecations and backwards incompatible changes. We recommend pinning any dependencies with `sweat==0.4.0`.**\n\nDocumentation can be found [here](https://sweatpy.gssns.io).\n\n## Contributors\n[Maksym Sladkov](https://github.com/sladkovm)\n[Aart Goossens](https://github.com/AartGoossens)\n\n## License\nSee [LICENSE](LICENSE) file.\n',
    'author': 'Aart Goossens',
    'author_email': 'aart@goossens.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/goldencheetah/sweatpy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



import pathlib

from fitparse import FitFile
import numpy as np
import pandas as pd

from .utils import remove_duplicate_indices, resample_data, semicircles_to_degrees


def read_fit(fpath, resample: bool = False, interpolate: bool = False) -> pd.DataFrame:
    """This method uses the Python fitparse library to load a FIT file into a Pandas DataFrame.
    It is tested with a Garmin FIT file but will probably work with other FIT files too.
    Columns names are translated to sweat terminology (e.g. "heart_rate" > "heartrate").

    Args:
        fpath: str, file-like or Path object
        resample: whether or not the data frame needs to be resampled to 1Hz
        interpolate: whether or not missing data in the data frame needs to be interpolated

    Returns:
        A pandas data frame with all the data.
    """

    if isinstance(fpath, pathlib.PurePath):
        fpath = fpath.as_posix()

    fitfile = FitFile(fpath)

    records = []
    lap = 0
    session = -1
    for record in fitfile.get_messages():
        if record.mesg_type is None:
            continue

        if record.mesg_type.name == "record":
            values = record.get_values()
            values["lap"] = lap
            values["session"] = session
            records.append(values)
        elif record.mesg_type.name == "lap":
            lap += 1
        elif record.mesg_type.name == "event":
            if record.get_value("event_type") == "start":
                # This happens whens an activity is (manually or automatically) paused or stopped and the resumed
                session += 1
        elif record.mesg_type.name == "sport":
            # @TODO handle this to be able to return metadata
            pass

    fit_df = pd.DataFrame(records)

    fit_df = fit_df.rename(
        columns={
            "heart_rate": "heartrate",
            "position_lat": "latitude",
            "position_long": "longitude",
            "altitude": "elevation",
            "left_right_balance": "left-right balance",
        }
    )

    fit_df["datetime"] = pd.to_datetime(fit_df["timestamp"], utc=True)
    fit_df = fit_df.drop(["timestamp"], axis="columns")
    fit_df = fit_df.set_index("datetime")

    for coordinate in ["latitude", "longitude"]:
        if coordinate in fit_df.columns:
            fit_df[coordinate] = semicircles_to_degrees(fit_df[coordinate])

    fit_df = remove_duplicate_indices(fit_df)

    fit_df = resample_data(fit_df, resample, interpolate)

    return fit_df



from pathlib import Path
from typing import Generator, Union

import pandas as pd

from .fit import read_fit
from .gpx import read_gpx
from .tcx import read_tcx


def read_file(fpath: Union[str, Path], *args, **kwargs) -> pd.DataFrame:
    """This method tries to recognize the file type of the fpath argument by reading the file extension (suffix).
    Please note that this method does not support file-like objects, in contrast to the other read_* functions of sweatpy.

    Args:
        fpath: str or Path object representing the path to a file.

    Returns:
        Returns an activity as a pandas data frames.
    """
    suffix = Path(fpath).suffix.lower()

    if suffix == ".tcx":
        read_func = read_tcx
    elif suffix == ".gpx":
        read_func = read_gpx
    elif suffix == ".fit":
        read_func = read_fit
    else:
        raise ValueError(
            f"Argument fpath ({fpath}) has an unsupported file extensions (suffix): {suffix}"
        )

    return read_func(fpath, *args, **kwargs)


def read_dir(path: Union[str, Path]) -> Generator[pd.DataFrame, None, None]:
    """Generator function that returns activities in a directory as pandas data frames.

    Args:
        path: str or Path object representing the path to a directory with activity files.

    Yields:
        Yields activities as pandas data frames.

    """
    path = Path(path)

    assert path.is_dir()

    for f in path.iterdir():
        if f.is_dir():
            continue

        yield read_file(f)



import pandas as pd


def resample_data(data, resample: bool, interpolate: bool) -> pd.DataFrame:
    """Function to calculate the mean max power for all available activities.

    Args:
        data: The data frame that needs to be resampled and/or interpolated
        resample: whether or not the data frame needs to be resampled to 1Hz
        interpolate: whether or not missing data in the data frame needs to be interpolated

    Returns:
        Returns the resampled and interpolated dataframe

    """
    if resample:
        data = data.resample("1S").mean()

    if interpolate:
        data = data.interpolate(method="linear")

    return data


def remove_duplicate_indices(data: pd.DataFrame, keep="first") -> pd.DataFrame:
    """Function that removes duplicate indices

    Args:
        data: The data frame for which duplicate indices need to be deleted
        keep: Determines which duplicates (if any) to mark. See pandas.DataFrame.index.duplicated documentation for more information

    Returns:
        Returns the data frame with duplicate indices removed

    """
    return data[~data.index.duplicated(keep=keep)]


def semicircles_to_degrees(semicircles):
    return semicircles * (180 / (2 ** 31))
